// TODO: Put public facing types in this file.

// import 'dart:async';
// import 'dart:io';

import 'package:kenken/kenken.dart';

/// Checks if you are awesome. Spoiler: you are.
class Kenken {
  Kenken(this.size) : cells = List.generate(size, (_) => List.filled(size, 0));

  /// The size of kenken. F.e. size=3 means that kenken has 3 rows and 3 columns.
  int size;

  /// Store for cell values, implemented as `List`(rows) of `List`(columns) of `int`(values).
  var cells = <List<int>>[];

  /// Store for all kenken rules;
  List<Rule> rules;

  /// Generates a kenken rules about rows and cols unique.
  List<Rule> get rowColRules {
    var rules = List.generate(size, (i) {
      var rowNumber = i + 1;
      var rowCoordinates = coordinatesOfRow(rowNumber);
      return Rule(
          rowCoordinates,
          Operation(isUniq, label: 'uniqueness in row', showLabel: false),
          ShouldHaveResult(true));
    });
    var colRules = List.generate(size, (i) {
      var colNumber = i + 1;
      var colCoordinates = coordinatesOfCol(colNumber);
      return Rule(
          colCoordinates,
          Operation(isUniq, label: 'uniqueness in col', showLabel: false),
          ShouldHaveResult(true));
    });
    rules.addAll(colRules);
    return rules;
  }

  /// Use it for set cell `value` by cell coordinates `x` and `y`.
  /// Coordinates should be in `1...kenken.size` diapazone.
  void setCell(int row, int col, int value) {
    var errMsg = _coordErrorMessage(row, col);
    if (errMsg.isNotEmpty) {
      throw ArgumentError(errMsg);
    }
    (cells[row - 1])[col - 1] = value;
  }

  /// Returns list of coordinates for row with given `rowNumber`
  List<CellWithRC> coordinatesOfRow(int rowNumber) =>
      List.generate(size, (i) => CellWithRC(rowNumber, i + 1));

  /// Returns list of coordinates for column with given `colNumber`
  List<CellWithRC> coordinatesOfCol(int colNumber) =>
      List.generate(size, (i) => CellWithRC(i + 1, colNumber));

  String _coordErrorMessage(int x, int y) {
    if (x < 1 || x > size || y < 1 || y > size) {
      return 'Bad coordinates: $x,$y';
    }
    return '';
  }

  /// Returns cell value
  int valueOf(CellWithRC cellCoord) {
    var errMsg = _coordErrorMessage(cellCoord.row, cellCoord.col);
    if (errMsg.isNotEmpty) {
      throw ArgumentError(errMsg);
    }
    return cells[cellCoord.row - 1][cellCoord.col - 1];
  }

  /// Returns invalid rules for kenken.
  /// If invalidRules list is empty, then kenken are solved.
  List<Rule> get invalidRules => rules.where((rule) {
        var ruleCellsValues =
            rule.cellsCoordinates.map((coords) => valueOf(coords)).toList();
        var actualResult = rule.operation.predicate(ruleCellsValues);
        return actualResult.value != rule.validOperationResult.value;
      }).toList();
}

/// A rule validates list of cells.
/// List of rules validates a whole kenken.
class Rule {
  Rule(this.cellsCoordinates, this.operation, this.validOperationResult);

  List<CellWithRC> cellsCoordinates;
  Operation operation;
  ShouldHaveResult validOperationResult;
  @override
  String toString() =>
      '$operation on $cellsCoordinates should be $validOperationResult';
}

/// Cell coordinates. Coordinates should be in 1...kenken.size diapazone.
class CellWithRC {
  CellWithRC(this.row, this.col);
  int row;
  int col;
  @override
  String toString() => '($row,$col)';
}

class ShouldHaveResult<T> {
  ShouldHaveResult(this.value);
  T value;
  @override
  String toString() => '$value';
}

/// Operation will be apply to list of cells for validate a rule.
/// Operation has predicate (function to be apply)
/// and printable operation symbol, accesible by `toString()` method (for display it on kenken).
class Operation {
  Operation(this.predicate, {this.label = '', this.showLabel = true});

  String label;
  Predicate predicate;
  bool showLabel;

  @override
  String toString() => label;
}

/// Function which can be applyed to List of values
typedef Predicate = ShouldHaveResult Function(List<int>);

/// Is list of values unique?
ShouldHaveResult isUniq(List<int> values) => ShouldHaveResult(values.isUniq);

/// Returns single value (first element value)
ShouldHaveResult singleValue(List<int> values) => ShouldHaveResult(values[0]);

/// Returns sum of elements
ShouldHaveResult sum(List<int> values) =>
    ShouldHaveResult(values.fold(0, (v, e) => v + e));

/// Returns difference of two values (biggest - smallest)
ShouldHaveResult diff(List<int> values) {
  if (values[0] > values[1]) {
    return ShouldHaveResult(values[0] - values[1]);
  }
  return ShouldHaveResult(values[1] - values[0]);
}

extension ListUniq<T> on List<T> {
  /// Is List values uniq?
  bool get isUniq {
    if (length < 2) return true;
    for (var i = 0; i < length - 1; i++) {
      if (sublist(i + 1).contains(this[i])) return false;
    }
    return true;
  }
}
